PKL_DEF_KEYWORD("asm")
PKL_DEF_KEYWORD("pinned")
PKL_DEF_KEYWORD("struct")
PKL_DEF_KEYWORD("union")
PKL_DEF_KEYWORD("else")
PKL_DEF_KEYWORD("while")
PKL_DEF_KEYWORD("until")
PKL_DEF_KEYWORD("for")
PKL_DEF_KEYWORD("in")
PKL_DEF_KEYWORD("where")
PKL_DEF_KEYWORD("if")
PKL_DEF_KEYWORD("sizeof")
PKL_DEF_KEYWORD("typeof")
PKL_DEF_KEYWORD("apush")
PKL_DEF_KEYWORD("apop")
PKL_DEF_KEYWORD("fun")
PKL_DEF_KEYWORD("method")
PKL_DEF_KEYWORD("type")
PKL_DEF_KEYWORD("var")
PKL_DEF_KEYWORD("unit")
PKL_DEF_KEYWORD("break")
PKL_DEF_KEYWORD("continue")
PKL_DEF_KEYWORD("return")
PKL_DEF_KEYWORD("string")
PKL_DEF_KEYWORD("as")
PKL_DEF_KEYWORD("try")
PKL_DEF_KEYWORD("catch")
PKL_DEF_KEYWORD("raise")
PKL_DEF_KEYWORD("void")
PKL_DEF_KEYWORD("any")
PKL_DEF_KEYWORD("print")
PKL_DEF_KEYWORD("printf")
PKL_DEF_KEYWORD("isa")
PKL_DEF_KEYWORD("unmap")
PKL_DEF_KEYWORD("remap")
PKL_DEF_KEYWORD("big")
PKL_DEF_KEYWORD("little")
PKL_DEF_KEYWORD("load")
PKL_DEF_KEYWORD("lambda")
PKL_DEF_KEYWORD("assert")
PKL_DEF_KEYWORD("format")
PKL_DEF_KEYWORD("enum")
PKL_DEF_KEYWORD("computed")
PKL_DEF_KEYWORD("immutable")
