/* Jitter: VM-independent library.

   Copyright (C) 2017 Luca Saiu
   Updated in 2021 by Luca Saiu
   Written by Luca Saiu

   This file is part of GNU Jitter.

   GNU Jitter is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   GNU Jitter is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Jitter.  If not, see <https://www.gnu.org/licenses/>. */


#include <stdbool.h>

#include "jitter/jitter.h"


/* Make sure that we provide the expected dispatch.
 * ************************************************************************** */

/* The name of this global depends on the dispatch, and is referred in
   the code generated by jitterc.  If the generated code is linked with the
   wrong runtime the mismatch will cause a link failure, instead of a misterious
   runtime crash. */
volatile const bool
JITTER_DISPATCH_DEPENDENT_GLOBAL_NAME = true;
