/* Uninspired VM: default -*- CSS -*- style for Libtextstyle.

   Copyright (C) 2020 Luca Saiu
   Written by Luca Saiu

   This file is part of the Uninspired VM implementation, distributed as
   an example along with GNU Jitter under the same license.

   GNU Jitter is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   GNU Jitter is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with GNU Jitter.  If not, see <https://www.gnu.org/licenses/>. */


/* Introduction.
 * ************************************************************************** */

/* This is a default style for the Uninspired VM, reasonable for terminals with
   a black background.  It looks much better with at least 256 colors, which is
   supported in practice by most modern terminals even when they do not set the
   TERM variable accordingly: the user is invited to try
     TERM=xterm-256color
   . */




/* Uninspired routines.
 * ************************************************************************** */

.uninspired-comment      { color: brown;
                           font-style: italic; }
.uninspired-punctuation  { color: white;  }
.uninspired-instruction  { color: cyan;
                           font-weight: bold; }
.uninspired-label        { color: yellow;
                           font-style: italic;
                           font-weight: bold; }
.uninspired-register     { color: green;
                           font-style: italic;
                           font-weight: bold; }
.uninspired-number       { color: lightgreen; }




/* Native-code disassemblying of Uninspired routines.
 * ************************************************************************** */

.uninspired-memory-address          { color: yellow;
                                      font-style: italic; }
.uninspired-native-instruction-hex  { color: cyan; }
.uninspired-disassembly             { color: green;
                                      font-style: italic; }




/* Interaction and messages.
 * ************************************************************************** */

.uninspired-warning  { background-color: red;
                       color: black;
                       font-style: italic;
                       text-decoration: underline; }




/* Profiling.
 * ************************************************************************** */

.uninspired-profile-instruction      { color: cyan;
                                       font-weight: bold; }
.uninspired-profile-execution-count  { color: brown; }
.uninspired-profile-run-time         { color: lightgreen;
                                       font-style: italic; }
.uninspired-profile-time-ratio       { color: gray;
                                       font-weight: bold; }
